type Address = Int

type Enviroment = (Memory, Binding)

type Binding = [Bind]
type Memory = [Cell]

type Bind = (Address, Id)
type Cell = (Address, Valor)

getNextPos :: Memory -> Address
getNextPos [] = 0
getNextPos (a:tail) = 1 + getNextPos tail

addBind :: Binding -> Address -> Id -> Binding
addBind ([]) (ad) (id) | (ad >= 0) = [(ad, id)]
                       | otherwise = []
addBind (a:tail) (ad) (id) | (ad < 0) = (a:tail)
addBind ((a, i):tail) (ad) (id) | (ad == a)     = (a, id):tail
                                | otherwise     = (a, i):(addBind tail ad id)

addCell :: Memory -> Address -> Valor -> Memory
addCell ([]) (ad) (val) | (ad >= 0) = [(ad, val)]   -- se a cauda acabou adiciona a nova celula
                        | otherwise = []
addCell (a:tail) (ad) (val) | (ad < 0) = (a:tail)
addCell ((a, v):tail) (ad) (val) | (ad == a)     = (a, val):tail     --sobreescrever um endereco ja existente
                                 | otherwise     = (a, v):(addCell tail ad val)  --verifica a proxima celula

getBind :: Binding -> Id -> Address
getBind [] _ = -1
getBind ((a, i):tail) id | i == id      = a
                         | otherwise    = getBind tail id

getValue :: Memory -> Address -> Valor
getValue [] _ = V3 "Voce colocou algo na Memoria, por acaso?! _"
getValue ((a, v):tail) ad | a == ad          = v
                          | otherwise        = getValue tail ad
                          
showBind :: Bind -> String
showBind (ad, id) = "(" ++ (show ad) ++ "," ++ id ++ ")"

findBind :: Cell -> Enviroment -> String
findBind _ (mem,[]) = "" -- Bind vazio - Nao achou o Binding correspondente - WRITE ou READ
findBind (ad,val) ((a,v):tail,(e,i):cauda) | (ad == a && ad == e) = showBind ((e,i))
                                           | otherwise = findBind (ad,val) ((a,v):tail,cauda)



printBindings :: Binding -> String
printBindings ([]) = "."
printBindings (a:tail) = (showBind a) ++ " , " ++ (printBindings tail)

showCell :: Cell -> String
showCell (ad, val) = "(" ++ (show ad) ++ "," ++ (show val) ++ ")"

printMemory :: Memory -> String
printMemory ([]) = "."
printMemory (a:tail) = (showCell a) ++ " , " ++ (printMemory tail)

printEnviroment :: Enviroment -> String
printEnviroment ([],_) = "."
printEnviroment (m:tail,[]) = "[" ++ showCell m ++ "]" ++ (printEnviroment (tail, []))
printEnviroment (m:tail, bin) = "[" ++ showCell m ++ " " ++ (findBind m (m:tail,bin)) ++ "]" ++
                                    (printEnviroment (tail, bin))

data Programa = PROG Comando

data Comando = ATRIB Id Expressao
               | DECL Declaracao Comando
               | WHILE Expressao Comando
               | IF_THEN_ELSE Expressao Comando Comando
               | CIO EntraSai
               | SEQ_COMM Comando Comando
               | SKIP
               | PROC Id ListaExpressao
               deriving Show


data Expressao = VAL Valor

               -- ExpUnaria
               | MINUS Expressao
               | NOT Expressao
               | LENGTH Expressao

               -- ExpBinaria
               | SOMA Expressao Expressao
               | SUBTRACAO Expressao Expressao
               | AND Expressao Expressao
               | OR Expressao Expressao
               | IGUAL Expressao Expressao
               | CONCATENA Expressao Expressao

               | IDENT Id
               | END_PONT Id
               deriving Show

data Declaracao = VAR_IGUAL Id Expressao
                | PONTEIRO Id Tipo
                | DECL_PROC DeclaracaoProc
                | VIRG Declaracao Declaracao
                deriving Show

data DeclaracaoProc = PROC_PAR Id ListaDeclaracaoParametro Comando
                    | PROC_VOID Id Comando
                    deriving Show

data ListaDeclaracaoParametro = SINGLE_DECL Tipo Id
                              | LIST_DECL Tipo Id ListaDeclaracaoParametro
                              deriving Show

data Tipo = TString
          | TInt
          | TBool
          deriving Show

data EntraSai = WRITE Expressao
              | READ Id
              deriving Show

data ListaExpressao = FINAL Expressao
                    | LIST_EXP Expressao ListaExpressao
                    | VAZIA
                    deriving Show

type Id = String

data Valor = V1 Int
           | V2 Bool
           | V3 String
           | PROCED DeclaracaoProc
           | Nil
           deriving Show

-- AQUI COMEA OS EVAL
interprete :: Programa -> Enviroment -> String
interprete (PROG prog) (env) = printEnviroment (evalComando (prog) env)

evalExpressao :: Expressao -> Enviroment -> Valor
evalExpressao (VAL v) (env) = v

evalExpressao (NOT e) (mem, bin) = opnot (evalExpressao e (mem, bin))
evalExpressao (MINUS e) (mem, bin) = opminus (evalExpressao e (mem, bin))
evalExpressao (LENGTH e) (mem, bin) = oplength (evalExpressao e (mem, bin))

evalExpressao (SOMA e1 e2) (mem, bin) = opsoma (evalExpressao e1 (mem, bin)) (evalExpressao e2 (mem, bin))
evalExpressao (SUBTRACAO e1 e2) (mem, bin) = opsubtracao (evalExpressao e1 (mem,bin)) (evalExpressao e2 (mem, bin))
evalExpressao (AND e1 e2) (mem, bin) = opand (evalExpressao e1 (mem, bin)) (evalExpressao e2 (mem, bin))
evalExpressao (OR e1 e2) (mem, bin) = opor (evalExpressao e1 (mem, bin)) (evalExpressao e2 (mem, bin))
evalExpressao (IGUAL e1 e2) (mem, bin) = V2 (opequal (evalExpressao e1 (mem, bin)) (evalExpressao e2 (mem, bin)))
evalExpressao (CONCATENA e1 e2) (mem, bin) = opconcat (evalExpressao e1 (mem, bin)) (evalExpressao e2 (mem, bin))

evalExpressao (IDENT i) (mem, bin) = getValue mem (getBind bin i)
-- r := &v
evalExpressao (END_PONT i) (mem, bin) = V1 (getBind bin i)

evalComando :: Comando -> Enviroment -> Enviroment
evalComando (ATRIB ident expr) (mem,bin) = evalAtribuicao ident expr (mem,bin)
evalComando (DECL decl comm) (mem,bin) = let env2 = evalDeclaracao decl (mem,bin) in (evalComando comm env2)
evalComando (WHILE expr comm) (mem,bin) = evalWhile expr comm (mem,bin)
evalComando (IF_THEN_ELSE expr comm1 comm2) (mem,bin) = evalIfThenElse expr comm1 comm2 (mem,bin)
evalComando (CIO entrada) (mem,bin) = evalIO entrada (mem,bin)
evalComando (SEQ_COMM comm1 comm2) (mem,bin) = evalSeqComm (comm1) (comm2) (mem,bin)
evalComando (SKIP) (mem,bin) = (mem,bin)
evalComando (PROC ident lista) (mem,bin) = evalProcedure ident lista (mem,bin)

evalAtribuicao :: Id -> Expressao -> Enviroment -> Enviroment
evalAtribuicao id expr (mem,bin) = (addCell mem (getBind bin id) (evalExpressao expr (mem,bin)), bin)

evalSeqComm :: Comando -> Comando -> Enviroment -> Enviroment
evalSeqComm comm1 comm2 (mem,bin) = let env2 = evalComando comm1 (mem,bin) in (evalComando comm2 env2)

evalDeclaracao :: Declaracao -> Enviroment -> Enviroment
evalDeclaracao (VAR_IGUAL i exp) (mem,bin)
                                = let pos = (getNextPos mem) in (addCell mem pos (evalExpressao exp (mem, bin)), addBind bin pos i)
evalDeclaracao (PONTEIRO i tipo) (mem,bin)
                                = let pos = (getNextPos mem) in (addCell mem pos Nil, addBind bin pos i)
evalDeclaracao (DECL_PROC dec) (mem,bin) = evalDeclaracaoProc dec (mem,bin)
evalDeclaracao (VIRG dec1 dec2) (mem,bin) = let env2 = evalDeclaracao dec1 (mem,bin) in (evalDeclaracao dec2 env2)

evalDeclaracaoProc :: DeclaracaoProc -> Enviroment -> Enviroment
evalDeclaracaoProc (PROC_PAR id listaArgs comm) (mem,bin)
       = let pos = (getNextPos mem) in (addCell mem pos (PROCED (PROC_PAR id listaArgs comm)), addBind bin pos id)
evalDeclaracaoProc (PROC_VOID id comm) (mem,bin)
       = let pos = (getNextPos mem) in (addCell mem pos (PROCED (PROC_VOID id comm)), addBind bin pos id)

evalProcedure :: Id -> ListaExpressao -> Enviroment -> Enviroment
evalProcedure (id) (VAZIA) (mem,bin) = evalComando (getComando (getValue mem (getBind bin id))) (mem,bin)
evalProcedure (id) (lista) (mem,bin)
              = let env2 = ligarParam (mem,bin) (getLista (getValue mem (getBind bin id))) lista  -- (getValue mem (getBind bin id)) = DeclaracaoProc
                in evalComando (getComando (getValue mem (getBind bin id))) env2

ligarParam :: Enviroment -> ListaDeclaracaoParametro -> ListaExpressao -> Enviroment
ligarParam (mem, bin) (SINGLE_DECL tp id) (FINAL exp)
           = let pos = (getNextPos mem) in (addCell (mem) pos (evalExpressao exp (mem, bin)), addBind (bin) (pos) (id))
ligarParam (mem, bin) (LIST_DECL tp id lista) (LIST_EXP exp le)
           = ligarParam (ligarParam (mem, bin) (SINGLE_DECL tp id) (FINAL exp)) lista le

getComando :: Valor -> Comando
getComando (PROCED (PROC_PAR id lista comm)) = comm
getComando (PROCED (PROC_VOID id comm)) = comm

getLista :: Valor -> ListaDeclaracaoParametro
getLista (PROCED (PROC_PAR id lista comm)) = lista

evalWhile :: Expressao -> Comando -> Enviroment -> Enviroment
evalWhile expr comm env | opequal (V2 True) (evalExpressao expr env)
                                             = let env2 = (evalComando comm env) in (evalWhile expr comm env2)
                        | otherwise          = env

evalIfThenElse :: Expressao -> Comando -> Comando -> Enviroment -> Enviroment
evalIfThenElse expr comm1 comm2 env | opequal (V2 True) (evalExpressao expr env)
                                                     = evalComando comm1 env
                                    | otherwise      = evalComando comm2 env

evalIO :: EntraSai -> Enviroment -> Enviroment
evalIO (WRITE exp) (mem,bin) = (addCell mem (getNextPos mem) (evalExpressao exp (mem,bin)), bin)
evalIO (READ id) (mem,bin) = (addCell mem (getNextPos mem) (getValue mem (getBind bin id)), bin)

showValor :: Valor -> String
showValor (V1 i) = show i
showValor (V2 b) = show b
showValor (V3 s) = show s

opnot :: Valor -> Valor
opnot (V2 b) = V2 (not b)

opminus :: Valor -> Valor
opminus (V1 i) = V1 (-i)

oplength :: Valor -> Valor
oplength (V3 s) = V1 (length s)

opsoma :: Valor -> Valor -> Valor
opsoma (V1 i) (V1 j) = V1 (i + j)

opsubtracao :: Valor -> Valor -> Valor
opsubtracao (V1 i) (V1 j) = V1 (i - j)

opand :: Valor -> Valor -> Valor
opand (V2 b1) (V2 b2) = V2 (b1 && b2)

opor :: Valor -> Valor -> Valor
opor (V2 b1) (V2 b2) = V2 (b1 || b2)

opequal :: Valor -> Valor -> Bool
opequal (V2 b1) (V2 b2) = (b1 == b2)
opequal (V1 i) (V1 j) = (i == j)
opequal (V3 s) (V3 t) = (s == t)

opconcat :: Valor -> Valor -> Valor
opconcat (V3 s) (V3 t) = V3 (s ++ t)

-- Exemplos de Codigo
-- int a = 5;
-- write(&a);
-- String b = "qualquer coisa";

saida = interprete (PROG (DECL (VAR_IGUAL "a" (VAL (V1 5))) (SEQ_COMM (CIO (WRITE (END_PONT "a"))) (DECL (VAR_IGUAL "b" (VAL (V3 "qualquer coisa"))) (SKIP)))))  ([],[])


